@@ -28,16 +28,44 @@ from . import fmaps
 from . import dcm2niix as d2n
 from .io import (read_json,
                  write_json,
-                 parse_bids_keyvals,
-                 parse_dcm2niix_fname,
+                 parse_bids_fname_keyvals,
                  safe_copy,
                  create_file_if_missing,
-                 strip_extensions)
+                 nii_to_json)
 
 
-def purpose_handling(bids_purpose,
+def add_participant_record(studydir, subject, age, sex):
+    """
+    Copied from heudiconv, this solution is good b/c it checks if the same subject ID already exists
+    :param studydir:
+    :param subject:
+    :param age:
+    :param sex:
+    :return:
+    """
+
+    participants_tsv = os.path.join(studydir, 'participants.tsv')
+    participant_id = 'sub-%s' % subject
+
+    if not create_file_if_missing(participants_tsv, '\t'.join(['participant_id', 'age', 'sex', 'handedness']) + '\n'):
+
+        # Check if subject record already exists
+        with open(participants_tsv) as f:
+            f.readline()
+            known_subjects = {this_line.split('\t')[0] for this_line in f.readlines()}
+
+        if participant_id in known_subjects:
+            return
+
+    # Add a new participant
+    with open(participants_tsv, 'a') as f:
+        f.write(
+            '\t'.join(map(str, [participant_id, age.lstrip('0').rstrip('Y') if age else 'N/A', sex, 'left'])) + '\n')
+
+
+def purpose_handling(bids_meta,
+                     bids_purpose,
                      bids_intendedfor,
-                     seq_name,
                      work_nii_fname,
                      work_json_fname,
                      bids_nii_fname,
@@ -48,10 +76,11 @@ def purpose_handling(bids_purpose,
     """
     Special handling for each image purpose (func, anat, fmap, dwi, etc)
 
+    :param bids_meta: dict
+        Metadata dict populated from BIDS JSON sidecar
     :param bids_purpose: str
         BIDS purpose directory name (eg anat, func, fmap, etc)
     :param bids_intendedfor: list of str
-    :param seq_name: str
     :param work_nii_fname: str
         work directory dcm2niix output Nifit filename
     :param work_json_fname: str
@@ -73,12 +102,12 @@ def purpose_handling(bids_purpose,
     bids_bval_fname = []
     bids_bvec_fname = []
 
-    # Load the JSON sidecar
-    bids_info = read_json(work_json_fname)
+    # Extract some useful fields from the BIDS metadata
+    scan_seq = bids_meta['ScanningSequence']
 
     if bids_purpose == 'func':
 
-        if 'EP' in seq_name:
+        if 'EP' in scan_seq:
 
             print('    EPI detected')
 
@@ -94,24 +123,24 @@ def purpose_handling(bids_purpose,
             create_events_template(bids_nii_fname, overwrite, nii_ext)
 
             # Add taskname to BIDS JSON sidecar
-            bids_keys = parse_bids_keyvals(bids_nii_fname)
+            bids_keys = parse_bids_fname_keyvals(bids_nii_fname)
             if 'task' in bids_keys:
-                bids_info['TaskName'] = bids_keys['task']
+                bids_meta['TaskName'] = bids_keys['task']
             else:
-                bids_info['TaskName'] = 'unknown'
+                bids_meta['TaskName'] = 'unknown'
 
     elif bids_purpose == 'fmap':
 
         # Add IntendedFor field if requested through protocol translator
         if 'UNASSIGNED' not in bids_intendedfor:
-            bids_info['IntendedFor'] = bids_intendedfor
+            bids_meta['IntendedFor'] = bids_intendedfor
 
         # Check for GRE vs SE-EPI fieldmap images
         # GRE will have a 'GR' sequence, SE-EPI will have 'EP'
 
         print('    Identifying fieldmap image type')
 
-        if seq_name == 'GR':
+        if 'GR' in scan_seq:
 
             print('    Gradient echo fieldmap detected')
             print('    Identifying magnitude and phase images')
@@ -119,7 +148,7 @@ def purpose_handling(bids_purpose,
             # Update BIDS filenames according to BIDS Fieldmap Case (1 or 2 - see specification)
             bids_nii_fname, bids_json_fname = fmaps.handle_fmap_case(work_json_fname, bids_nii_fname, bids_json_fname)
 
-        elif seq_name == 'EP':
+        elif 'EP' in scan_seq:
 
             print('    EPI fieldmap detected')
 
@@ -134,15 +163,15 @@ def purpose_handling(bids_purpose,
 
     elif bids_purpose == 'anat':
 
-        if seq_name == 'GR_IR':
+        if 'GR' in scan_seq and 'IR' in scan_seq:
 
             print('    IR-prepared GRE detected - likely T1w MPRAGE or MEMPRAGE')
 
-            # Handle multiecho EPI (echo-*). Modify bids fnames as needed
+            # Handle MEMPRAGE. Modify bids fnames as needed
             bids_nii_fname, bids_json_fname = d2n.handle_multiecho(
                 work_json_fname, bids_json_fname, key_flags['Echo'], nii_ext)
 
-            # Handle complex-valued EPI (part-*). Modify bids fnames as needed
+            # Handle complex-valued MEMPRAGE. Modify bids fnames as needed
             bids_nii_fname, bids_json_fname = d2n.handle_complex(
                 work_json_fname, bids_json_fname, key_flags['Part'], nii_ext)
 
@@ -150,13 +179,13 @@ def purpose_handling(bids_purpose,
             bids_nii_fname, bids_json_fname = d2n.handle_bias_recon(
                 work_json_fname, bids_json_fname, key_flags['Recon'], nii_ext)
 
-        elif seq_name == 'SE':
+        elif 'SE' in scan_seq:
 
             print('    Spin echo detected - likely T1w or T2w anatomic image')
             bids_nii_fname, bids_json_fname = d2n.handle_bias_recon(
                 work_json_fname, bids_json_fname, key_flags['Recon'], nii_ext)
 
-        elif seq_name == 'GR':
+        elif 'GR' in scan_seq:
 
             print('    Gradient echo detected')
 
@@ -176,7 +205,7 @@ def purpose_handling(bids_purpose,
         safe_copy(work_nii_fname, str(bids_nii_fname), overwrite)
 
     if bids_json_fname:
-        write_json(bids_json_fname, bids_info, overwrite)
+        write_json(bids_json_fname, bids_meta, overwrite)
 
     if bids_bval_fname:
         safe_copy(work_bval_fname, bids_bval_fname, overwrite)
@@ -185,50 +214,21 @@ def purpose_handling(bids_purpose,
         safe_copy(work_bvec_fname, bids_bvec_fname, overwrite)
 
 
-def add_participant_record(studydir, subject, age, sex):
-    """
-    Copied from heudiconv, this solution is good b/c it checks if the same subject ID already exists
-    :param studydir:
-    :param subject:
-    :param age:
-    :param sex:
-    :return:
+def add_run_number(bids_stub, run_no):
     """
+    Safely add run number to BIDS stub
+    Allow protocol translator to override auto run numbering
 
-    participants_tsv = os.path.join(studydir, 'participants.tsv')
-    participant_id = 'sub-%s' % subject
-
-    if not create_file_if_missing(participants_tsv, '\t'.join(['participant_id', 'age', 'sex', 'group']) + '\n'):
-
-        # Check if subject record already exists
-        with open(participants_tsv) as f:
-            f.readline()
-            known_subjects = {this_line.split('\t')[0] for this_line in f.readlines()}
-
-        if participant_id in known_subjects:
-            return
-
-    # Add a new participant
-    with open(participants_tsv, 'a') as f:
-        f.write(
-            '\t'.join(map(str, [participant_id, age.lstrip('0').rstrip('Y') if age else 'N/A', sex, 'control'])) + '\n')
-
-
-def add_run_number(bids_suffix, run_no):
-    """
-    Safely add run number to BIDS suffix
-    Handle prior existence of run-* in BIDS filename template from protocol translator
-
-    :param bids_suffix, str
+    :param bids_stub, str
     :param run_no, int
-    :return: new_bids_suffix, str
+    :return: new_bids_stub, str
     """
 
-    # Init return value
-    new_bids_suffix = bids_suffix
+    # Default returns original stub unchanged
+    new_bids_stub = bids_stub
 
     # Extract BIDS keys from suffix
-    bids_keys, _ = parse_bids_keyvals(bids_suffix)
+    bids_keys, _ = parse_bids_fname_keyvals(bids_stub)
 
     if 'run' in bids_keys.keys():
 
@@ -240,9 +240,9 @@ def add_run_number(bids_suffix, run_no):
         # Skip adding run number if series is singular (run_no < 0)
         if run_no > 0:
             bids_keys['run'] = run_no
-            new_bids_suffix = bids_keys_to_filename(bids_keys, '')
+            new_bids_stub = bids_keys_to_filename(bids_keys, '')
 
-    return new_bids_suffix
+    return new_bids_stub
 
 
 def add_bids_key(bids_json_fname, key_name, key_value, nii_ext):
@@ -280,7 +280,7 @@ def bids_filename_to_keys(bids_fname):
     """
 
     # Parse BIDS filename with internal function that supports part- key
-    keys, dname = parse_bids_keyvals(bids_fname)
+    keys, dname = parse_bids_fname_keyvals(bids_fname)
 
     # Substitute short key names
     if 'subject' in keys:
@@ -315,7 +315,12 @@ def bids_keys_to_filename(keys, dname):
 
     # Add final pulse sequence suffix and extension
     if 'suffix' in keys:
-        bids_fname += keys['suffix']
+        if len(keys['suffix']) > 0:
+            bids_fname += keys['suffix']
+
+    # Remove final trailing '_' (no suffix case)
+    if bids_fname[-1] == '_':
+        bids_fname = bids_fname[:-1]
 
     if 'extension' in keys:
         bids_fname += keys['extension']
@@ -339,38 +344,45 @@ def bids_legalize_keys(keys):
     return keys
 
 
-def auto_run_no(file_list, prot_dict):
+def auto_run_no(d2n_nii_list, prot_dict):
     """
     Search for duplicate series names in dcm2niix output file list
     Return inferred run numbers accounting for duplication and multiple recons from single acquisition
 
     NOTES:
-    - Multiple recons generated by single acquisition (eg multiecho fieldmaps, localizers, etc) are
+    - Multiple recons generated by single acquisition (eg multiecho mag/phase fieldmaps, localizers, etc) are
       handled through the dcm2niix extensions (_e1, e2_ph, _i00001, etc).
-    - Series number resets following subject re-landmarking make the SerNo useful only for
+    - DICOM SeriesNo resets following subject re-landmarking make the SerNo useful only for
       determining series uniqueness and not for ordering or run numbering.
     - If no duplicates of a given series are found, drop the run- key from the BIDS filename
     - Current dcm2niix version: 1.0.20211006
 
-    :param file_list: list of str
-        Nifti file name list
+    :param d2n_nii_list: list of str
+        dcm2niix output Nifti filename list
     :param prot_dict: dictionary
         Protocol translation dictionary
     :return: run_num, array of int
     """
 
     # Construct list of series descriptions and original numbers from file names
-    desc_list = []
+    series_id_list = []
 
-    for fname in file_list:
+    # Loop over all
+    for nii_fname in d2n_nii_list:
 
-        # Parse dcm2niix filename into relevant keys, including suffix
-        info = parse_dcm2niix_fname(fname)
+        # Load JSON sidecar for this Nifti image
+        json_fname = nii_to_json(nii_fname, '.nii.gz')
+        bids_info = read_json(json_fname)
 
-        ser_desc = info['SerDesc']
+        ser_desc = bids_info['SeriesDescription'].replace(' ', '_')
+        if 'EchoNumber' in bids_info.keys():
+            echo_no = bids_info['EchoNumber']
+        else:
+            echo_no = 1
+        recon_type = '-'.join(bids_info['ImageType'])
 
         if ser_desc in prot_dict:
-            _, bids_suffix, _ = prot_dict[info['SerDesc']]
+            _, bids_stub, _ = prot_dict[ser_desc]
         else:
             print('')
             print('* Series description {} missing from code/Protocol_Translator.json'.format(ser_desc))
@@ -378,29 +390,29 @@ def auto_run_no(file_list, prot_dict):
             print('* Exiting')
             sys.exit(1)
 
-        # Construct a unique series description using multirecon suffix
-        ser_suffix = bids_suffix + '_' + info['Suffix']
+        # Construct a unique series identifier including echo number and suffix
+        series_id = f"{bids_stub}_ECHO{echo_no}_{recon_type}"
 
         # Add to list
-        desc_list.append(ser_suffix)
+        series_id_list.append(series_id)
 
     # Find unique ser_desc entries using sets
-    unique_descs = set(desc_list)
+    unique_series_ids = set(series_id_list)
 
     # Init vector of run numbers and max run numbers for each series
-    run_no = np.zeros(len(file_list)).astype(int)
+    run_no = np.zeros(len(d2n_nii_list)).astype(int)
 
     # Loop over unique series descriptions
-    for unique_desc in unique_descs:
+    for unique_series_id in unique_series_ids:
 
         # Count duplicates of unique description in description list
-        n_dup = desc_list.count(unique_desc)
+        n_dup = series_id_list.count(unique_series_id)
 
         if n_dup == 1:
 
             # Replace run number of singular series with -1 to indicate
             # that run- should be dropped in BIDS filename creation
-            run_no[desc_list.index(unique_desc)] = -1
+            run_no[series_id_list.index(unique_series_id)] = -1
 
         else:
 
@@ -408,16 +420,16 @@ def auto_run_no(file_list, prot_dict):
             run_count = 0
 
             # Loop over all series descriptions
-            for i, desc in enumerate(desc_list):
+            for i, desc in enumerate(series_id_list):
 
-                if desc == unique_desc:
+                if desc == unique_series_id:
                     run_count += 1
                     run_no[i] = run_count
 
     return run_no
 
 
-def replace_contrast(fname, new_contrast):
+def replace_suffix(fname, new_contrast):
     """
     Replace contrast suffix (if any) of BIDS filename with new_contrast
 
@@ -426,7 +438,7 @@ def replace_contrast(fname, new_contrast):
     :return: new_fname: str, modified BIDS filename
     """
 
-    bids_keys, dname = parse_bids_keyvals(fname)
+    bids_keys, dname = parse_bids_fname_keyvals(fname)
     bids_keys['suffix'] = new_contrast
     new_fname = bids_keys_to_filename(bids_keys, dname)
 
@@ -475,15 +487,21 @@ def create_events_template(bold_fname, overwrite, nii_ext):
             fd.close()
 
 
-def auto_translate(info, json_fname):
+def auto_translate(info, json_fname=None):
     """
     Construct protocol translator from original series descriptions
-    - supports ReproIn-style series descriptions with a leading "seqtype-" key
+    - supports ReproIn-style series descriptions with a leading "<BIDS type>-<suffix>" key
+      eg anat-T1w_acq-highres_run-1
+
+    :param info: dict
+        Series information including SeriesDescription
+    :param json_fname: str, pathlike
+        Path of JSON sidecar in working directory [optional]
     """
 
-    ser_desc = info['SerDesc']
+    ser_desc = info['SeriesDescription']
 
-    # List of possible suffices for each BIDS type directory
+    # List of possible suffices for each BIDS type
     bids_types = {
         'func': ['bold', 'sbref'],
         'anat': ['T1w', 'T2w', 'PDw', 'T2starw', 'FLAIR',
@@ -495,15 +513,41 @@ def auto_translate(info, json_fname):
 
     # Use BIDS filename parser on ReproIn-style series description
     # Returns any BIDS-like key values from series description string
-    # The closer the series descriptions are to Repro-In specs, the
+    # The closer the series descriptions are to ReproIn specs, the
     # better this works.
-    bids_keys, _ = parse_bids_keyvals(ser_desc)
+    bids_keys, _ = parse_bids_fname_keyvals(ser_desc)
 
-    # Give precedence to ReproIn seqtype key
     if 'seqtype' in bids_keys.keys():
-
         bids_dir = bids_keys['seqtype']
-        print(f"ReproIn: detected sequence type {bids_dir}")
+        print(f"Detected legacy seqtype {bids_dir}")
+
+    elif 'anat' in bids_keys.keys():
+
+        # ReproIn anat-<sequence type> prefix
+        bids_dir = 'anat'
+        bids_keys['suffix'] = bids_keys['anat']
+        del bids_keys['anat']
+
+    elif 'func' in bids_keys.keys():
+
+        # ReproIn func-<sequence type> prefix
+        bids_dir = 'func'
+        bids_keys['suffix'] = bids_keys['func']
+        del bids_keys['func']
+
+    elif 'fmap' in bids_keys:
+
+        # ReproIn fmap-<sequence type> prefix
+        bids_dir = 'fmap'
+        bids_keys['suffix'] = bids_keys['fmap']
+        del bids_keys['fmap']
+
+    elif 'dwi' in bids_keys:
+
+        # ReproIn dwi-<sequence type> prefix
+        bids_dir = 'dwi'
+        bids_keys['suffix'] = bids_keys['dwi']
+        del bids_keys['dwi']
 
     else:
 
@@ -523,10 +567,9 @@ def auto_translate(info, json_fname):
     # Filled automatically if requested during translation
     bids_intendedfor = 'UNASSIGNED'
 
-    print('')
-    print(f'Series Description : {ser_desc}')
-    print(f'BIDS directory     : {bids_dir}')
-    print(f'BIDS stub          : {bids_stub}')
-    print(f'BIDS IntendedFor   : {bids_intendedfor}')
+    print(f'  Series Description : {ser_desc}')
+    print(f'  BIDS directory     : {bids_dir}')
+    print(f'  BIDS stub          : {bids_stub}')
+    print(f'  BIDS IntendedFor   : {bids_intendedfor}')
 
     return [bids_dir, bids_stub, bids_intendedfor]
\ No newline at end of file
